#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

from ..base.psu.simple_psu import SimplePsu
from ..korad.korad_slow_serial import KoradSlowSerial

class LABPS3005DN(SimplePsu):
    ''' This is based on the interface for the Korad KAxxxx power supplies.
        Unfortunately, the commands are slightly different.
        The handling of the serial data seems to be the same, so the Korad serial interface can be used

        Note that the LABPS3005D is identical to the Korad KA5003, but the 'DN' is not.
    '''

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)

        self.link = KoradSlowSerial(self, baudrate=9600)
    
    def _set_enabled(self, value:bool):
        with self.link as lnk:
            lnk.write("OUTPUT1\\n" if value else "OUTPUT0\\n")

    def _get_enabled(self) -> bool:
        raise NotImplementedError()

    def _set_current(self, value:float):
        value = max(0,min(value,5))

        with self.link as lnk:
            lnk.write(f"ISET1:{value:05.3f}\\n")

    def _get_current(self) -> float:
        with self.link as lnk:
            return lnk.query_float("ISET1?\\n")

    def _read_current(self) -> float:
        with self.link as lnk:
            return lnk.query_float("IOUT1?\\n")

    def _set_voltage(self, value:float):
        value = max(0,min(value,30))
            
        with self.link as lnk:
            lnk.write(f"VSET1:{value:05.2f}\\n")

    def _get_voltage(self) -> float:
        with self.link as lnk:
            return lnk.query_float("VSET1?\\n")

    def _read_voltage(self) -> float:
        with self.link as lnk:
            return lnk.query_float("VOUT1?\\n")



#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

class Attribute:
    ''' Like a property for a class, but with extra functionality.
        Intended to be used with classes derived from Instrument_Base
        Calls the method _cache_value on the object to remember the value that
        was last read or written.
    '''

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

    def __set_name__(self, owner, name):
        self.__name__ = name

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        if self.fget is None:
            raise AttributeError( f'No setter for {type(obj).__name__!r}.{self.__name__!r}' )
        value = self.fget(obj)
        obj._cache_value(self.__name__, value)
        return value

    def __set__(self, obj, value):
        if self.fset is None:
            raise AttributeError
        obj._cache_value(self.__name__, value)
        self.fset(obj, value)

    def __delete__(self, obj):
        if self.fdel is None:
            raise AttributeError
        self.fdel(obj)

    def getter(self, fget):
        return type(self)(fget, self.fset, self.fdel, self.__doc__)

    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel, self.__doc__)

    def deleter(self, fdel):
        return type(self)(self.fget, self.fset, fdel, self.__doc__)
    

class AttributeRef():
    ''' Reference to a property/Attribute of a class.
        Is returned when using [] on an object with the property name.
        Use methods get() and set() to get/set the property value.
        When used in combination with InstrumentBase, also supports methods peek() and poke()
        to access values using caching to avoid unnecesasay commands to the instrument.
    '''

    def __init__(self, obj:'SubSystem', propname):
        # if not hasattr(obj, propname):                    # this does a get !!!
        #     raise KeyError(f"Class '{type(obj).__name__}' has no property '{propname}'")
        self.obj = obj
        self.propname = propname

    def set(self, value):
        setattr( self.obj, self.propname, value)

    def get(self):
        return getattr( self.obj, self.propname)
    
    def peek(self):
        return self.obj.peek(self.propname)
    
    def poke(self, value):
        return self.obj.poke(self.propname, value)
    

class AttributeProvider():
    ''' Base class for subsystems of an instrument.
    '''
    def __init__(self):
        self.__attribute_cache = dict()               # for caching Attribute values

    def __getitem__(self, propname):
        return AttributeRef(self, propname)

    def _cache_value(self, name, value):
        ''' Called by setter/getter of an Attribute.  The cached value is used by
            peek() and poke() to avoid unnecessary read and writes to the instrument.
        '''
        self.__attribute_cache[name] = value

    def peek(self, name):
        ''' Return the cached value of an Attribute.
            If no value is cached yet, the actual value of the Attributed is requested.
        '''
        if name in self.__attribute_cache:
            return self.__attribute_cache[name]
        else:
            return getattr(self, name)
    
    def poke(self, name, value):
        ''' Write a value to an Attribute using caching.
            It is ignored if the new value is the same as the cached value.
        '''
        cache_value = self.__attribute_cache.get(name)
        if cache_value is None or cache_value!=value:
            setattr(self, name, value)



#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

from ..base.psu.simple_psu import SimplePsu

class PSI9750_06DT_Simple(SimplePsu):
    ''' Base class to derive instruments from with a given voltage/current range.
    '''

    def __init__(self, name, **kwargs):
        super().__init__(name, terminator="lf", **kwargs)

        from ..links import Visa
        self.link = Visa(self)

    def to_float(self, txt):
        parts = txt.split()
        if len(parts)==0:
            raise ValueError(f"Received unexpected reply '{txt}'")
        try:
            return float(parts[0])
        except:
            raise ValueError(f"Reply '{txt}' doesn't contain a number")

    def lock(self):
        with self.link as lnk:
            lnk.write("SYST:LOCK ON")

    def unlock(self):
        with self.link as lnk:
            lnk.write("SYST:LOCK OFF")

    def _set_enabled(self, value:bool):
        with self.link as lnk:
            lnk.write("OUTP 1" if value else "OUTP 0")

    def _get_enabled(self):
        with self.link as lnk:
            status = lnk.query("OUTP?")
            return True if status=="ON" else False

    def _set_current(self, value:float):
        with self.link as lnk:
            lnk.write(f"CURR {value:0.3f}")

    def _get_current(self) -> float:
        with self.link as lnk:
            return self.to_float(lnk.query("CURR?"))

    def _read_current(self) -> float:
        with self.link as lnk:
            return self.to_float(lnk.query("MEAS:CURRENT?"))

    def _set_voltage(self, value:float):
        with self.link as lnk:
            lnk.write(f"VOLT {value:0.1f}")

    def _get_voltage(self) -> float:
        with self.link as lnk:
            return self.to_float(lnk.query("VOLT?"))

    def _read_voltage(self) -> float:
        with self.link as lnk:
            return self.to_float(lnk.query("MEAS:VOLT?"))




#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

import threading
from .instrument import Instrument
from .config import Config

class LinkBase():
    
    def __init__(self, instrument : Instrument, **kwargs):
        '''
        '''
        self._inst = instrument
        self.log = instrument._log                                      # take logger object from instrument
        self._config = Config.get_config(instrument._name, **kwargs)    # take settings from ini-file and add kwargs

        self._lock = threading.Lock()                       # semaphore for access to this link

    ######## open and close ###################################################
                        
    def open(self):
        ''' Open the port if necessary.  Does nothing if port is already open.
        '''
        pass
    
    def close(self):
        pass

    def __del__(self):
        ''' Close all connections (if necessary) when connection is destroyed
            (typically when the program is terminated).
        '''
        self.close()

    ######## getting and releasing access #####################################

    def acquire(self):
        ''' Acquires semaphore for using the connection.
            Then return an object with all the interface methods using get_link().
        '''
        self._lock.acquire()
        self.open()
        return self

    def release(self):
        ''' Releases the semaphore for using the connection.
            Must be called after every acquire()
        '''
        self._lock.release()

    def __enter__(self):
        ''' Automatic acquire/release of semaphore using context manager
        '''
        return self.acquire()
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        ''' Automatic acquire/release of semaphore using context manager.
        '''
        self.release()
        return False


    
        




#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0


from ..subsystem import SubSystem
from ..attribute import Attribute

class OscilloscopeChannel(SubSystem):
    ''' Interface class to return actual voltage and current.
    '''
    
    def __init__(self, inst, index):
        super().__init__(inst)
        self._inst = inst
        self._index = index

    @property
    def _link(self):
        return self._inst._link

    def enable(self):
        self.enabled = True

    def disable(self):
        self.enabled = False

    @Attribute
    def enabled(self):
        return self._get_enabled()

    @enabled.setter
    def enabled(self, value):
        self._set_enabled(value)

    @Attribute
    def scale(self):
        return self._get_scale()
    
    @scale.setter
    def scale(self, value):
        self._set_scale(value)

    @Attribute
    def offset(self):
        return self._get_offset()

    @offset.setter
    def offset(self, value):
        self._set_offset(value)

    @Attribute
    def coupling(self):
        return self._get_coupling()

    @coupling.setter
    def coupling(self, value):
        self._set_coupling(value)

    @Attribute
    def bandwidth(self):
        return self._get_bandwidth()

    @bandwidth.setter
    def bandwidth(self, value):
        self._set_bandwidth(value)

    @Attribute
    def impedance(self):
        return self._get_impedance()

    @impedance.setter
    def impedance(self, value):
        self._set_impedance(value)

    # abstract methods below; to be implemented in derived class
    
    def _get_enabled(self):
        raise NotImplementedError()
    def _set_enabled(self, value:bool):
        raise NotImplementedError()

    def _get_scale(self):
        raise NotImplementedError()
    def _set_scale(self, value:float):
        raise NotImplementedError()

    def _get_offset(self):
        raise NotImplementedError()
    def _set_offset(self, value:float):
        raise NotImplementedError()

    def _get_coupling(self):
        raise NotImplementedError()
    def _set_coupling(self, value:bool):
        raise NotImplementedError()

    def _get_bandwidth(self):
        raise NotImplementedError()
    def _set_bandwidth(self, value:int):
        raise NotImplementedError()

    def _get_impedance(self):
        raise NotImplementedError()
    def _set_impedance(self, value:int):
        raise NotImplementedError()





#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

import numpy as np

from ...base.oscilloscope.oscilloscope import Oscilloscope
from .infiniium_channel import InfiniiumChannel


class InfiniiumScope(Oscilloscope):
    
    
    def __init__(self, name, **kwargs):
        
        super().__init__(name, **kwargs)

        from ...links import Visa
        self._link = Visa(self)

        # two-channel scope
        self.channels = [ InfiniiumChannel(self, idx) for idx in range(2) ]
                
    def get_id(self):
        ''' Return ID of scope
        '''
        with self._link as lnk:
            return lnk.query("*IDN?")
    



#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

from ..korad.kaxxxxp import KA3005P

class LABPS3005D(KA3005P):
    ''' LABPS3005D is identical to Korad KA5003.

        Note that the LABPS3005DN is not identical.  See LABPS3005DN.py
    '''

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)



#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

import numpy as np

from ...base.oscilloscope.oscilloscope_channel import OscilloscopeChannel
from ...base.oscilloscope.oscillocope_constants import OscilloscopeConstants as OCT

class InfiniiumChannel(OscilloscopeChannel):
    
    
    def __init__(self, inst, index):
        
        super().__init__(inst, index)

    @property
    def channel(self) -> int:
        return self._index+1

    def _get_enabled(self):
        raise NotImplementedError()
    
    def _set_enabled(self, visible:bool):
        with self._link as lnk:
            if visible:
                lnk.write("CHAN%d:DISP ON" % self.channel )
            else:
                lnk.write("CHAN%d:DISP OFF" % self.channel )        

    def _get_scale(self):
        raise NotImplementedError()
    
    def _set_scale(self, voltage:float):
        with self._link as lnk:
            lnk.write(":CHAN%d:SCAL %G" % (self.channel, voltage) )

    def _get_offset(self):
        raise NotImplementedError()
    
    def _set_offset(self, offset:float):
        with self._link as lnk:
            lnk.write(":CHAN%d:OFFS %G" % (self.channel, offset) )

    def _get_coupling(self):
        raise NotImplementedError()
    
    def _set_coupling(self, value):
        if value==OCT.coupling.AC:
            with self._link as lnk:
                lnk.write(":CHAN%d:PROB:COUP AC" % self.channel )
        elif value==OCT.coupling.DC:
            with self._link as lnk:
                lnk.write(":CHAN%d:PROB:COUP DC" % self.channel )
        else:
            raise ValueError(f"{value} is not valid for Infiniium channel coupling mode.")

    def _get_bandwidth(self):
        raise NotImplementedError()
    def _set_bandwidth(self, value:int):
        raise NotImplementedError()

    def _get_impedance(self):
        raise NotImplementedError()
    def _set_impedance(self, value:int):
        raise NotImplementedError()





#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0


import logging
from .attribute import AttributeProvider
from .config import Config

class Instrument(AttributeProvider):
    ''' Base class for all instruments
    '''
    
    def __init__(self, name, **kwargs):
        super().__init__()

        self._name = name
        self._config = Config.get_config(name, **kwargs)              # take settings from ini-file and kwargs
        
        self._log = logging.getLogger("inst.%s" % name)
        self._link = None

    # def get_config(self):
    #     ''' Returns the section of a ConfigParser with settings for this instrument.
    #     '''
    #     return self._config
    



#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

import time
import serial

from ..base.link_base import LinkBase
from ..base.instrument import Instrument


class KoradSlowSerial(LinkBase):
    ''' Handler for communication with Korad power supplies using the serial port.
        See https://sigrok.org/wiki/Korad_KAxxxxP_series#Protocol for details.

        Delays have been determined experimentally to avoid communication errors.
    '''

    WRITE_DELAY_S = 0.055                # minimum time between two writes
    READ_WAIT_S = 0.1                    # minimum time to wait before aborting a read

    def __init__(self, instrument:'Instrument', **kwargs):
        ''' Opens connection over serial port.
        '''
        super().__init__(instrument, **kwargs)
        
        self.port = None
        self.last_write = time.time()

    def open(self):
        ''' Opens serial port if necessary
        '''
        if self.port is not None:                               # skip if port already open
            return

        if "comport" in self._config:
            portname = self._config.get("comport")
        else:
            raise Exception(f"Comport not found in settings for {self._inst._name}")
        
        baudrate = self._config.getint("baudrate", fallback = 9600)

        self.log.debug(f"Opening COM port {portname} at {baudrate} baud")

        try:                                                    # open port with 20ms read timeout
            self.port = serial.Serial(portname, baudrate, timeout=0.02)
        except serial.serialutil.SerialException:
            self.log.error("Could not open COM port %s", portname)
            self.port = None

    def close(self):
        ''' Closes serial thread and connection.
        '''
        if self.port is None:                                   # skip if port not open
            return

        self.port.close()
        self.port = None

    def write(self, data:str):
        ''' Writes binary data to the port.
            Time is tracked to make sure that two writes are not too fast after one another.
        '''
        assert self.port is not None, "Serial port must be opened first"

        data = data.strip().encode()

        dt = time.time() - self.last_write

        if dt < self.WRITE_DELAY_S:                     # check if we didn't write too fast
            dt = self.WRITE_DELAY_S - dt                # the extra time we need to wait
            if dt < 0.005:
                time.sleep(0.005)                       # wait at least 5ms
            else:
                time.sleep(dt)                          # otherwise just wait the extra time

        self.port.write(data)                           # write data
        self.port.flush()
        # remember time of last write
        self.last_write = time.time()

    def read(self, numbytes=8):
        ''' Reads data, up to a number of bytes.
        '''
        rd_start = time.time()                          # remember when we start reading

        buf = bytearray()                               # buffer with result

        while True:
            in_byte = self.port.read(1)                 # read a byte

            if len(in_byte) == 0:                               # no data any more
                if time.time() - rd_start > self.READ_WAIT_S:   # and we've waited long enough
                    print("read timeout")
                    break                                       # then it's OK to stop
                else:
                    continue

            buf += bytearray(in_byte)                   # add received byte to buffer

            if len(buf) >= numbytes:                    # stop if enough bytes read
                break

        if len(buf) == 0:                               # no data received -> return None
            return None

        if len(buf) == 1:                               # single byte = status -> return as integer
            return int(buf[0])

        return buf.decode("utf-8")                      # >1 byte = text reply -> return as string

    def write_read(self, cmd, numbytes=99):
        ''' Sends command, wait 1/10th of a second, and then checks the reply.
            Stupid way to do because I couldn't get a queue or semaphore to work.
        '''
        self.port.reset_input_buffer()
        self.write(cmd)                                    # send command
        return self.read(numbytes)

    def query_float(self, cmd):
        ''' Sends a command that expects a floating point number as reply.
            Returns the floating point number.
            
            Korad always replies with exactly 5 characters for a number, either xx.xx or x.xxx.
            Reply can contain a leftover character from a previous command (firmware bug).
            Therefore only characters '.' and '1'-'9' are accepted.
            Any illegal character is assumed to be from a previous reply.
        '''
        if not '?' in cmd:
            raise ValueError("Korad query command must contain question mark.")
    
        self.write(cmd)
    
        rd_start = time.time()                          # remember when we start reading

        buf = bytearray()                               # buffer with result

        while len(buf)<5:                               # read at most 5 bytes
            
            in_byte = self.port.read(1)                 # read a byte

            if len(in_byte) == 0:                               # no data any more
                if time.time() - rd_start > self.READ_WAIT_S:   # and we've waited long enough
                    break                                       # then it's OK to stop
                else:
                    continue                                    # go wait for next character

            the_byte = in_byte[0]
            if the_byte!=46 and (the_byte<48 or the_byte>57):   # not a valid character
                buf = bytearray()                               # reset receive buffer
                continue

            buf += bytearray(in_byte)                   # add received byte to buffer

        # print(f"<<{buf.decode('utf-8')}<<")

        if len(buf)!=5:                                 # something wrong : reply must always be 5 bytes
            raise ValueError("Power supply didn't reply with 5 bytes.")

        try:
            fvalue = float(buf.decode("utf-8"))         # try to convert reply to floating point
        except (ValueError, TypeError):
            fvalue = -1
            
        if fvalue<0:
            raise ValueError("Power supply didn't return a valid value.")

        return fvalue



#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

import numpy as np

from .infiniium import InfiniiumScope, InfiniiumChannel


class DSO54830B(InfiniiumScope):
    
    def __init__(self, name, **kwargs):
        
        super().__init__(name, num_channels=2, **kwargs)





#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

from ...base.link_base import LinkBase
from ...base.instrument import Instrument

pyvisa_rm = None

class PyVisaLink(LinkBase):
    ''' 
    '''

    def __init__(self, instrument:'Instrument', **kwargs):
        ''' Opens connection over serial port.
        '''
        super().__init__(instrument, **kwargs)
        
        self._pyvisa_conn = None


    def acquire(self):
        ''' the context manager should return the pyvisa connection itself,
            not this wrapper class.
        '''
        super().acquire()
        return self._pyvisa_conn

    def open(self):
        ''' Opens pyvisa connection if necessary
        '''
        if self._pyvisa_conn is not None:
            return

        if not "pyvisa" in self._config:                                # check if the connection string is found
            raise Exception(f"visa address not found in settings for {self._inst._name}")

        global pyvisa_rm
        if pyvisa_rm is None:                                       # instantiate resource manager if necessary
            import pyvisa
            pyvisa_rm = pyvisa.ResourceManager()
        
        self._pyvisa_conn = pyvisa_rm.open_resource(self._config["pyvisa"])    # open the instrument
        
        if "baudrate" in self._config:
            self._pyvisa_conn.baud_rate = self._config.getint("baudrate")
        if "flowcontrol" in self._config:
            self._pyvisa_conn.flow_control = self._config.getint("flowcontrol")
        if "terminator" in self._config:
            eol = self._config.get("terminator", "crlf").lower()
            if eol=="cr":
                self._pyvisa_conn.read_termination='\r'
                self._pyvisa_conn.write_termination='\r'
            elif eol=="lf":
                self._pyvisa_conn.read_termination='\n'
                self._pyvisa_conn.write_termination='\n'
            elif eol=="crlf":
                self._pyvisa_conn.read_termination='\r\n'
                self._pyvisa_conn.write_termination='\r\n'
            else:
                raise Exception(f"value {eol} for visa_eol is invalid, excpect cr, lf or crlf")

    def close(self):
        ''' Closes connection to instrument
        '''
        if self._pyvisa_conn is None:
            return
        
        self._pyvisa_conn.close()
        self._pyvisa_conn = None




"""InstrumentLab - controlling test instruments like power supplies and oscilloscopes."""

__version__ = "0.0.5"

from .base.config import Config

from .base.oscilloscope.oscillocope_constants import OscilloscopeConstants



#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0


from ..instrument import Instrument
from ..subsystem import SubSystem
from ..attribute import Attribute

class SimplePsuReadout(SubSystem):
    ''' Interface class to return actual voltage and current.
    '''
    
    def __init__(self, inst:'SimplePsu'):
        super().__init__(inst)
        self._inst = inst
    
    @Attribute
    def voltage(self):
        return self._inst._read_voltage()

    @Attribute
    def current(self):
        return self._inst._read_current()


class SimplePsu(Instrument):
    ''' Interface class for the basic operations :
        * get and set current and voltage
        * enable/disable output + convenience functions
        * return actual output values
    '''
    
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)

        self.read = SimplePsuReadout(self)
    
    @Attribute
    def enabled(self):
        return self._get_enabled()

    @enabled.setter
    def enabled(self, value):
        self._set_enabled(value)   

    def enable(self):
        ''' Convencience function to enable the output.'''
        self._set_enabled(True)

    def disable(self):
        ''' Convenience function to disable the output.'''
        self._set_enabled(False)

    @Attribute
    def voltage(self):
        return self._get_voltage()

    @voltage.setter
    def voltage(self, value):
        self._set_voltage(value)   
    
    @Attribute
    def current(self):
        return self._get_current()

    @current.setter
    def current(self, value):
        self._set_current(value)

    # abstract methods below; to be implemented in derived class

    def _set_enabled(self, value:bool):
        raise NotImplementedError()

    def _get_enabled(self) -> bool:
        raise NotImplementedError()
    
    def _set_voltage(self, value:float):
        raise NotImplementedError()

    def _get_voltage(self) -> float:
        raise NotImplementedError()

    def _read_voltage(self) -> float:
        raise NotImplementedError()

    def _set_current(self, value:float):
        raise NotImplementedError()

    def _get_current(self) -> float:
        raise NotImplementedError()

    def _read_current(self) -> float:
        raise NotImplementedError()



#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

import numpy as np

from ..base.oscilloscope.oscilloscope import Oscilloscope
from ..base.subsystem import SubSystem

class WaveForm:
    
    def __init__(self, scope, channel):
        ''' Gets waveform of the specified channel.
        '''
        self.channel = channel
        self.data = scope.get_numpy_data(channel)
        (self.x0, self.dx, self.y0, self.dy) = scope.get_data_info()
        self.size = self.data.size
        

    def xvalues(self, scale=1):
        ''' Returns numpy array of X-values
        '''
        return np.linspace(self.x0*scale, (self.x0+self.dx*self.size)*scale, self.size)


class AgilentMeasure(SubSystem):
    
    def __init__(self, inst):
        super().__init__(inst)
        
    def fall_time(self, channel):
        cmd = ":MEAS:FALL? CHANNEL%d" % channel
        return self.get_value(cmd)

    def rise_time(self, channel):
        cmd = ":MEAS:RIS? CHANNEL%d" % channel
        return self.get_value(cmd)
    
    def vmax(self, channel):
        cmd = ":MEAS:VMAX? CHANNEL%d" % channel
        return self.get_value(cmd)

    def vmin(self, channel):
        cmd = ":MEAS:VMIN? CHANNEL%d" % channel
        return self.get_value(cmd)
    
    def vbase(self, channel):
        cmd = ":MEAS:VBAS? CHANNEL%d" % channel
        return self.get_value(cmd)
    
    def vtop(self, channel):
        cmd = ":MEAS:VTOP? CHANNEL%d" % channel
        return self.get_value(cmd)
    

class DSO54830B(Oscilloscope):
    
    
    def __init__(self, name, **kwargs):
        
        super().__init__(name, **kwargs)

        from ..links import Visa
        self._link = Visa(self)
        
        self.measure = AgilentMeasure(self)        
        
    def get_id(self):
        ''' Return ID of scope
        '''
        with self._link as lnk:
            return lnk.query("*IDN?")
    
    def set_defaults(self):
        ''' scope to default settings.
        '''
        self.write(":SYSTEM:PRESET")
        
    def set_acquisition_points(self, count):
        ''' Set number of acquisition points.
        '''
        self.write(":ACQ:INT OFF")                 # no sin(x0/x interpolation
        self.write(":ACQ:POIN %d" % count)         # set number of points
        self.write(":ACQ:SRAT AUTO")               # sample rate automatic
        
    def set_acquisition_auto(self):
        ''' Set acquisition points and rate to auto.
        '''
        self.write(":ACQ:INT OFF")
        self.write(":ACQ:POIN AUTO")
        self.write(":ACQ:SRAT AUTO")
        
    def acquire(self, timeout=5):
        ''' Make sure a new waveform is acquired.
        '''
        self.query(":ADER?")                       # clear ADER event
        
        count = 0
        while count < 10*timeout:
            reply = self.query(":ADER?") 
            if reply.startswith("+1"):
                break
        
    def set_timebase(self, time_per_div):
        ''' Sets the time per division for the timebase (horizontal)
        '''
        self.write(":TIM:SCAL %G" % time_per_div)
        
    def get_time_range(self):
        ''' Gets the time range of the window that is displayed.
            Result : left time, right time, full range, position
        '''
#         pos = self.get_value(":TIM:WIND:POS?")
#         rang = self.get_value(":TIM:WIND:RANG?")
        pos = self.get_value(":TIM:POS?")
        rang = self.get_value(":TIM:RANG?")
        
        left = -rang/2 + pos
        right = rang/2 + pos
        return (left, right, rang, pos)
        
        
    def set_trigger(self, channel, position=None, level=None, rising=None, falling=None):
        ''' Sets the trigger.
        '''
        self.write(":TRIG:EDGE:SOURCE CHAN%d" % channel)
        
        if position is not None:
            self.write(":TIM:POS %G" % position)
            
        if rising is not None or falling is not None:
            if rising == True and falling == True:
                self.write(":TRIG:EDGE:SLOP EITH")
            elif rising == True or falling == False:
                self.write(":TRIG:EDGE:SLOP POS")
            else:
                self.write(":TRIG:EDGE:SLOP NEG")

        if level is not None:
            self.write(":TRIG:LEV CHAN%d, %G" % (channel, level) )
        
    def set_vertical(self, channel, voltage, offset = None):
        ''' \public 
            Set voltage/div for channel, and optionally offset.
            Negative offset moves the waveform up.
        '''
        self.write(":CHAN%d:SCAL %G" % (channel, voltage) )
        if offset is None:
            offset = 3*voltage
        self.write(":CHAN%d:OFFS %G" % (channel, offset) )
        
    def show(self, channel, visible=True):
        ''' Show or hide a channel.
        '''
        if visible:
            self.write("CHAN%d:DISP ON" % channel )
        else:
            self.write("CHAN%d:DISP OFF" % channel )
            
    def get_data(self, channel):
        ''' Get data for a channel
        '''
        if isinstance(channel, str):
            self.write(":WAV:SOUR %s" % channel )
        else:
            self.write(":WAV:SOUR CHAN%d" % channel )
        self.write(":WAV:FORM ASC" )
        self.write(":WAV:VIEW ALL" )
        data = self.query_ascii_values(":WAV:DATA?")
        
        return data
    
    def get_numpy_data(self, channel):
        ''' Get data for a channel as a numpy array
        '''
        if isinstance(channel, str):
            self.write(":WAV:SOUR %s" % channel )
        else:
            self.write(":WAV:SOUR CHAN%d" % channel )
        self.write(":WAV:FORM ASC" )
        self.write(":WAV:VIEW ALL" )
        data = self.get_array(":WAV:DATA?")
        return data
    
    def get_data_info(self):
        ''' Get info about waveform data like start point, offset, ...
        '''
        x0 = self.get_value(":WAV:XOR?")
        dx = self.get_value(":WAV:XINC?")
        y0 = self.get_value(":WAV:YOR?")
        dy = self.get_value(":WAV:YINC?")
    
        return (x0, dx, y0, dy)
    
    def get_waveform(self, channel):
        ''' Returns an object with data and info.
        '''
        return WaveForm(self, channel)
    


#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

import configparser 
import os
    
class Config():
    
    config = None
    
    @classmethod
    def read(cls, filename:str, optional=False):
        
        if not filename.endswith(".ini"):
            raise ValueError(f"Filename {filename} must end with .ini")

        filename = os.path.abspath(filename)
        
        if not os.path.isfile(filename):
            if not optional:
                raise FileNotFoundError((f"Filename {filename} not found"))
        else:
            if cls.config is None:
                cls.config = configparser.ConfigParser()
                cls.config.read(filename)
            else:
                new_config = configparser.ConfigParser()
                new_config.read(filename)

                #TODO: add sections from this config file to the config that is already read
                raise NotImplementedError("reading multiple config files not yet supported")
        
    @classmethod
    def get_config(cls, name:str, **kwargs) -> configparser.SectionProxy:
        ''' Returns settings for the given section in the ini-files.
            
            The arguments supplied as kwargs is also included, as if they were defined in
            the ini-file itself.  If there are duplicates, the ini-file values have priority.
            
            Everything is returned as a new ConfigParser section containing all values.
        '''
        
        if cls.config is None:                                  # create empty configparser if nothing has ever been read
            cls.config = configparser.ConfigParser()

        if not cls.config.has_section(name):                    # make empty section if nothing for this name yet
            cls.config[name] = dict()

        section = cls.config[name]                              # get the section for this name

        for key,value in kwargs.items():                        # add parameters provided to constructor
            if key not in section:                              # but only if they don't exist yet
                section[key] = str(value)

        return section                                          # now return only the section for this name




#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

from ..base.psu.simple_psu import SimplePsu
from .korad_slow_serial import KoradSlowSerial

class KoradSimplePSU(SimplePsu):
    ''' Base class to derive instruments from with a given voltage/current range.
    '''

    def __init__(self, name, max_voltage, max_current, **kwargs):
        super().__init__(name, **kwargs)

        self.link = KoradSlowSerial(self)

        self.max_voltage = max_voltage
        self.max_current = max_current
    
    def _set_enabled(self, value:bool):
        with self.link as lnk:
            lnk.write("OUT1" if value else "OUT0")

    # TODO: implement _get_enabled()

    def _set_current(self, value:float):
        value = max(0,min(value,self.max_current))

        with self.link as lnk:
            lnk.write(f"ISET1:{value:05.3f}")

    def _get_current(self) -> float:
        with self.link as lnk:
            return lnk.query_float("ISET1?")

    def _read_current(self) -> float:
        with self.link as lnk:
            return lnk.query_float("IOUT1?") 

    def _set_voltage(self, value:float):
        value = max(0,min(value,self.max_voltage))
            
        with self.link as lnk:
            lnk.write(f"VSET1{value:05.2f}")

    def _get_voltage(self) -> float:
        with self.link as lnk:
            return lnk.query_float("VSET1?")

    def _read_voltage(self) -> float:
        with self.link as lnk:
            return lnk.query_float("VOUT1?")


class KA3005P(KoradSimplePSU):

    def __init__(self, name, **kwargs):
        super().__init__(name, 30, 5, **kwargs)

class KA6003P(KoradSimplePSU):

    def __init__(self, name, **kwargs):
        super().__init__(name, 60, 3, **kwargs)





# import tomllib
import os
import re

TOML_FILE = os.path.join(os.path.dirname(os.path.realpath(__file__)),"pyproject.toml")
INIT_FILE = os.path.join(os.path.dirname(os.path.realpath(__file__)),"instrumentlab","__init__.py")

new_version = None

def update_toml_file():
    ''' Open the toml file, find the version number, and increment it.
    '''
    global new_version
    
    with open(TOML_FILE, 'r') as rf:                # read entire toml file
        data = rf.readlines()
    
    with open(TOML_FILE, 'w') as wf:                # start rewriting it
        for line in data:                           # go over all the lines
            
            m = re.match('version = \"(\d*).(\d*).(\d*)\"', line)   # does it contain the version number ?
            if m:
                v1,v2,v3 = m.group(1),m.group(2),m.group(3)         # get all parts
                
                new_version = f"{v1}.{v2}.{int(v3)+1}"              # build new version number
    
                print( f"Update version from {v1}.{v2}.{v3} to {new_version}")
                
                wf.write(f'version = "{new_version}"\n')                 # write new version number

            else:
                wf.write(line)                      # just copy all other lines


def update_init_file():
    ''' Open the __init__ file of the package, and fill in the version number.
    '''
    with open(INIT_FILE, 'r') as rf:                # read entire init file
        data = rf.readlines()
    
    with open(INIT_FILE, 'w') as wf:                # start rewriting it
        for line in data:                           # go over all the lines
            
            if line.startswith("__version__"):      # write new version
                wf.write(f'__version__ = "{new_version}"\n')
            else:
                wf.write(line)                      # just copy all other lines


if __name__=="__main__":
    update_toml_file()
    assert new_version is not None, "Version number not found in toml file"
    update_init_file()
    
    
    
    


#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

from ...base.config import Config
from ...base.instrument import Instrument

def Visa(inst:'Instrument'):
    ''' Return a link to a Visa instrument.
        Typically, pyvisa is used.  However, other implementations could be possible.
    '''

    config = Config.get_config(inst._name)              # get the settings for the instrument

    if "pyvisa" in config:                              # we need to use pyvisa
        from .pyvisa_link import PyVisaLink
        return PyVisaLink(inst)                         # return the wrapper for pyvisa
    
    return ConnectionRefusedError( 
        f"No configuration settings found for Visa connection to {inst._name}")
    



#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0

import logging

# from .instrument import Instrument
from .attribute import AttributeProvider

class SubSystem(AttributeProvider):
    ''' Base class for subsystems of an instrument.
    '''
    def __init__(self, inst:'Instrument'):
        super().__init__()
        self._inst = inst
   
    @property
    def link(self):
        return self._inst.link
    
    @property
    def log(self) -> logging.Logger:
        return self._inst.log



#  ___         _                          _   _         _    
# |_ _|_ _  __| |_ _ _ _  _ _ __  ___ _ _| |_| |   __ _| |__   InstrumentLab
#  | || ' \(_-<  _| '_| || | '  \/ -_) ' \  _| |__/ _` | '_ \  
# |___|_||_/__/\__|_|  \_,_|_|_|_\___|_||_\__|____\__,_|_.__/  (C) 2024  Marc Van Riet et al.
#
# Licensed under the Apache License Version 2.0. See http://www.apache.org/licenses/LICENSE-2.0


from ..instrument import Instrument
from ..attribute import Attribute
from .oscilloscope_channel import OscilloscopeChannel



class Oscilloscope(Instrument):
    ''' Interface class for the basic operations :
        * get and set current and voltage
        * enable/disable output + convenience functions
        * return actual output values
    '''
    
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)

        self.channels = [ OscilloscopeChannel(self, 0) ]

    
 